package functionalDependency.general;

import java.io.StringReader;
import java.util.BitSet;
import java.util.Set;
import java.util.TreeSet;

import functionalDependency.parser.Parser;

/**
 * Die Klasse <code>AttributeSet</code> repraesentiert eine Menge von Attributen.
 * Die Attribute werden in der <code>BitSet</code>-Variablen <code>val</code>
 * durch gesetzte Bits dargestellt.
 * <p>Implementiert sind neben den Konstruktoren die Methoden</p>
 * <ul>
 * <li> {@link #parseSchemaFromLine(String)}: Parsen eines Strings
 * <li> {@link #add(char)}, {@link #add(String)}: Hinzufuegen von Attributen
 * (gegeben als <code>Character</code> oder <code>String</code>)
 * <li> {@link #remove(char)}: Loeschen von Attributen
 * <li> {@link #intersection(AttributeSet)}, {@link #union(AttributeSet)},
 * {@link #substract(AttributeSet)}, {@link #substract(char)}: Mengenoperationen
 * (Durchschnitt, Vereinigung, Differenz)
 * <li> {@link #includes(AttributeSet)}: Teilmengentest
 * <li> verschiedene Gettermethoden, Attributvergleichsmethoden,
 * eine Klon- und eine Ausgabemethode.
 * </ul>
 * <p>Im Rahmen der Aufgabenstellung muss in dieser Klasse nichts geaendert werden.</p>
 */
public class AttributeSet implements Cloneable,Comparable{
	
	protected BitSet val;
	
	public AttributeSet(){
		val = new BitSet();
	}
	
	public AttributeSet(String s) throws AttributeSetException{ 
		val = new BitSet();
		add(s);
	}

	public AttributeSet(char c) throws AttributeSetException{ 
		val = new BitSet();
		add(c);
	}
	
	private AttributeSet(BitSet v){
		val = (BitSet)v.clone();
	}
	
	static public AttributeSet parseSchemaFromLine(String line) throws MainException{
		Parser parser = null;
		StringReader reader;
		reader = new StringReader(line);
		parser = new Parser(reader);
		return parser.parseAttrDef();
	}

	public boolean add(String s) throws AttributeSetException{
		boolean rc=true;
		if(s.length()==0)
			throw new AttributeSetException("empty string",AttributeSetException.EMPTY_STR);
		for(int i=0;i<s.length();++i){
			rc &= add(s.charAt(i));
		}
		return rc;
	}

	public boolean add(char c) throws AttributeSetException{
		if(c>='A'&&c<='Z'){
			if (!val.get(c-'A')){
				val.set(c-'A');
				return true;
			}
		}else
			throw new AttributeSetException("invaild character",AttributeSetException.INVALID_CHARACTER);
		return false;
	}

	public boolean remove(char c) throws AttributeSetException{
		if(c>='A'&&c<='Z'){
			if (val.get(c-'A')){
				val.clear(c-'A');
				return true;
			}
		}else
			throw new AttributeSetException("invaild character",AttributeSetException.INVALID_CHARACTER);
		return false;
	}
	
	public boolean isEmpty(){
		return val.length()==0;
	}
	
	public int size(){
		int size = 0;
		for(int i=0;i<val.size();++i){
			if(val.get(i))
				++size;
		}
		return size;
	}

	public AttributeSet intersection(AttributeSet a){
		AttributeSet rc = this.clone();
		rc.val.and(a.getBitSet());
		return rc;
	}

	public AttributeSet union(AttributeSet a){
		AttributeSet rc = this.clone();
		rc.val.or(a.getBitSet());
		return rc;
	}
	
	/** substracts a from this **/
	public AttributeSet substract(AttributeSet a){
		AttributeSet rc = this.clone();
		rc.val.andNot(a.getBitSet());
		return rc;
	}

	/** substracts c from this **/
	public AttributeSet substract(char c) throws AttributeSetException{
		return substract(new AttributeSet(c));
	}

	/** true if a is part of this **/
	public boolean includes(AttributeSet a){
		BitSet tmp = (BitSet)val.clone();
		tmp.and(a.getBitSet());
		return tmp.equals(a.getBitSet());
	}
	
	/** true if this contains c */
	public boolean contains(char c) {
		return val.get(c - 'A');
	}
	
	private BitSet getBitSet(){
		return val;
	}
	
	public String getAttrString(){
		String rc="";
		char c = 'A';
		for(int i=0;i<val.size();++i,++c){
			if(val.get(i))
				rc += c;
		}
		return rc;
	}
	
	public Set<Character> getSet(){
		TreeSet<Character> rc = new TreeSet<Character>();
		char c = 'A';
		for(int i=0;i<val.size();++i,++c){
			if(val.get(i))
				rc.add(c);
		}
		return rc;
	}
	
	public boolean equals(AttributeSet a){
		return val.equals(a.getBitSet());
	}

	public boolean equals(Object a){
		if(a instanceof AttributeSet)
			return equals((AttributeSet)a);
		return false;
	}
	
	public int compareTo(Object o){
		if(o instanceof AttributeSet)
			return compareTo((AttributeSet)o);
		return -1;
	}
	
	public int compareTo(AttributeSet a){
		return toString().compareTo(a.toString());
	}

	public AttributeSet clone(){
		return new AttributeSet(this.getBitSet());
	}
	
	public String toString(){
		return "{" + getAttrString() + "}";
	}
	
	/**
	 * Transferiert eine Attributmenge in ein Array.
	 * 
	 * @return Array der Attribute
	 */
	public char[] toArray() {
		char[] attrchar = new char[this.size()];
		int arraycounter = 0;
		for (char c = 'A'; c <='Z'; c++) {
			if (this.contains(c)) {
				attrchar[arraycounter] = c;
				arraycounter++;
			}
		}
		return attrchar;
	}
}
