package functionalDependency.general;

/**
 * Die Klasse <code>FunctionalDependency</code> repraesentiert eine funktionale
 * Abhaengigkeit. Diese wird in den beiden <code>AttributeSet</code>-Variablen
 * <code>dominator</code> und <code>isDominated</code> gespeichert.
 * Beispielsweise gilt fuer eine FD <code>X --> Y</code>:
 * <code>dominator = X</code> und <code>isDominated = Y</code>.
 * <p>Implementiert sind neben den Konstruktoren die Methoden</p>
 * <ul>
 * <li> {@link #getAttributes()}: berechnet die Attributmenge aus der
 * <code>dominator</code>- vereinigt mit der <code>isDominated</code>-Attributmenge.
 * <li> verschiedene Gettermethoden, Vergleichsmethoden,
 * eine Klon- und eine Ausgabemethode.
 * </ul>
 * <p>Im Rahmen der Aufgabenstellung muss in dieser Klasse nichts geaendert werden.</p>
 */
public class FunctionalDependency implements Cloneable,Comparable{
	
	private AttributeSet dominator;
	private AttributeSet isDominated;
	
	private FunctionalDependency() {
		dominator = new AttributeSet();
		isDominated = new AttributeSet();		
	}

	public FunctionalDependency(String dominatorStr,
								String dominatedStr) throws MainException{
		dominator = new AttributeSet();
		isDominated = new AttributeSet();
		if(dominatorStr.length()==0)
			throw new FunctionalDependencyException("empty left side",FunctionalDependencyException.EMPTY_LEFT_SIDE);
		if(dominatedStr.length()==0)
			throw new FunctionalDependencyException("empty right side",FunctionalDependencyException.EMPTY_RIGHT_SIDE);
		dominator.add(dominatorStr);
		isDominated.add(dominatedStr);
	}

	public FunctionalDependency(AttributeSet dominatorSet,
			char dominatedChar) throws MainException{
		dominator = dominatorSet.clone();
		isDominated = new AttributeSet(dominatedChar);
	}

	public FunctionalDependency(AttributeSet dominatorSet,
			AttributeSet dominatedSet) throws MainException{
		dominator = dominatorSet.clone();
		isDominated = dominatedSet.clone();
	}
	
	public boolean equals(Object fd){
		if(fd instanceof FunctionalDependency)
			return equals((FunctionalDependency)fd);
		return false;
	}
	
	public boolean equals(FunctionalDependency fd){
		if(dominator.equals(fd.getDominator()) && isDominated.equals(fd.getDominated())){
			return true;
		}
		return false;
	}
	
	public int compareTo(Object o){
		if(o instanceof FunctionalDependency)
			return compareTo((FunctionalDependency)o);
		return -1;
	}
	
	public int compareTo(FunctionalDependency f){
		return toString().compareTo(f.toString());
	}
	
	public FunctionalDependency clone(){
		FunctionalDependency rc = new FunctionalDependency();
		rc.dominator = dominator.clone();
		rc.isDominated = isDominated.clone();
		return rc;
	}
	
	/**
	 * Berechnet die Menge von Attributen, die im aktuellen Objekt (FD) vorkommen.
	 * @return Menge von Attributen, die in der aktuellen funktionalen Abhaengigkeit
	 * (<code>this</code>) vorkommen.
	 */
	public AttributeSet getAttributes() {
		AttributeSet attr = new AttributeSet();
		attr = attr.union(this.getDominated()).union(this.getDominator());
		return attr;
	}
	
	public AttributeSet getDominator(){
		return dominator;
	}

	public AttributeSet getDominated(){
		return isDominated;
	}
	
	public String toString(){
		return dominator.getAttrString() + "->" + isDominated.getAttrString(); 
	}
}
