package functionalDependency.general;

import java.io.StringReader;
import java.util.Iterator;

import functionalDependency.parser.Parser;

/**
 * Die Klasse <code>RelationSchema</code> repraesentiert ein Relationsschema.
 * In den privaten Variablen <code>name</code>, <code>attributes</code>,
 * <code>keys</code> und <code>normalForm</code> werden der Name des
 * Schemas, die Attributmenge, die Schluesselkandidaten und die Normalform
 * gespeichert. Die Normalform kann die Werte <code>FirstNF</code>,
 * <code>SecondNF</code>, <code>ThirdNF</code> oder <code>BCNF</code> annehmen.
 * <p>Implementiert sind neben den Konstruktoren die Methoden</p>
 * <ul>
 * <li> {@link #parseSchemaFromLine(String, String)}: Parsen eines Strings
 * <li> verschiedene Gettermethoden, Vergleichsmethoden, eine Klon- und eine
 * Ausgabemethode.
 * </ul>
 * <p>Im Rahmen der Aufgabenstellung sind in dieser Klasse die Methoden</p>
 * <ul>
 * <li> {@link #getPrimeAttributeSet()}: Berechnen der Schluesselattribute
 * <b>(2 Punkte)</b>
 * <li> {@link #getNonPrimeAttributeSet()}: Berechnen der Nichtschluesselattribute
 * <b>(2 Punkte)</b>
 * <li> {@link #genNormalForm(SetOfFDs)}: Berechnen der Normalform
 * <b>(5 Punkte)</b>
 * </ul>
 * zu implementieren.
 */
public class RelationSchema implements Cloneable,Comparable {
	
	private String name;
	private AttributeSet attributes;
	private SetOfAttributeSets keys;
	private int normalForm;
	
	public static final int FirstNF = 1;
	public static final int SecondNF = 2;
	public static final int ThirdNF = 3;
	public static final int BCNF = 4;

	
	private RelationSchema() {
		name = null;
		attributes = null;
		keys = null;
		normalForm = 0;
	}
	
	public RelationSchema(String nm, String attr, SetOfFDs fds) throws MainException{
		name = nm;
		attributes = new AttributeSet (attr);
		keys = fds.genKeys(attributes);
		normalForm = genNormalForm(fds);
	}

	public RelationSchema(String nm, AttributeSet attr, SetOfFDs fds) throws MainException{
		name = nm;
		attributes = attr.clone();
		keys = fds.genKeys(attributes);
		normalForm = genNormalForm(fds);
	}
	
	static public RelationSchema parseSchemaFromLine(String name,String line) throws MainException{
		Parser parser = null;
		StringReader reader;
		reader = new StringReader(line);
		parser = new Parser(reader);
		return parser.parseRelationSchema(name);
	}
	
	public String getName() {
		return name;
	}

	public SetOfAttributeSets getKeys() {
		return keys;
	}
	
	public AttributeSet getAttributes() {
		return attributes;
	}
	
	public int getNormalForm() {
		return normalForm;
	}
	
	public boolean equals(Object r){
		if(r instanceof RelationSchema)
			return equals((RelationSchema)r);
		return false;
	}

	/**
	 * Vergleicht das Relationsschema <code>this</code> mit einem anderen
	 * Relationsschema <code>rs</code>.
	 * @param rs das zu vergleichende Relationsschema
	 * @return <code>true</code>, wenn die Relationsschemata gleich sind;
	 * <code>false</code>, andernfalls
	 */
	public boolean equals(RelationSchema rs){
		if (/*name.equalsIgnoreCase(rs.getName()) &&*/
				attributes.equals(rs.getAttributes()) &&
				keys.equals(rs.getKeys()) &&
				normalForm == rs.getNormalForm()) {
			return true;
		}
		return false;
	}
	
	public int compareTo(Object o){
		if(o instanceof RelationSchema)
			return compareTo((RelationSchema)o);
		return -1;
	}
	
	/**
	 * Vergleich zweier Relationsschemata.
	 * @param a das zu vergleichende Relationsschema
	 * @return 
	 */
	public int compareTo(RelationSchema a){
		return toStringCompare().compareTo(a.toStringCompare());
	}

	public RelationSchema clone(){
		RelationSchema rc = new RelationSchema();
		rc.name = new String(name);
		rc.attributes = attributes.clone();
		rc.keys = keys.clone();
		rc.normalForm = normalForm;
		return rc;
	}

	public String toString(){
		return name + "=" + attributes + " Keys=" + getKeys()
			+ " NormalForm=" + getNormalForm();
	}
	
	/**
	 * Ausgabe des Relationschemas ohne den Namen.
	 * Mithilfe dieses Strings laesst sich ein Vergleich
	 * (siehe {@link #compareTo(RelationSchema)})
	 * zweier Relationsschemata herstellen.
	 * @return ein <code>String</code> bestehend aus den Attributen,
	 * Schluesselkandidaten und der Normalform des Relationsschemas.
	 */
	public String toStringCompare(){
		return attributes.toString() + getKeys().toString() + getNormalForm();
	}

	
//	Studentenversion
//	----------------

	/**
	 * Berechnet die Schluesselattribute des gegebenen Relationsschema.
	 * Man beachte, dass die Schluesselkandidaten im Relationsschema bereits
	 * in <code>keys</code> gespeichert sind.
	 * @return Menge der Schluesselattribute
	 */
	public AttributeSet getPrimeAttributeSet() {
		SetOfAttributeSets schluessel = keys.clone();
		AttributeSet result = new AttributeSet();
		while (!schluessel.isEmpty()) {
			result = result.union(schluessel.first());
			schluessel.remove(schluessel.first());
		}
		return result;
	}

	/**
	 * Berechnet die Nichtschluesselattribute des gegebenen Relationsschema.
	 * Man beachte, dass die Schluesselkandidaten im Relationsschema bereits
	 * in <code>keys</code> gespeichert sind.
	 * @return Menge der Nichtschluesselattribute
	 */
	public AttributeSet getNonPrimeAttributeSet() {
		AttributeSet prime = getPrimeAttributeSet();
		AttributeSet nonprime = attributes.substract(prime);
		return nonprime;
	}
	
	/**
	 * Berechnet die Normalform des Relationsschemas.
	 * @param fds die FDs des Relationsschemas
	 * @return Normalform ({@link #FirstNF}, {@link #SecondNF}, 
	 * {@link #ThirdNF} oder {@link #BCNF}).
	 */
	private int genNormalForm(SetOfFDs fds) throws MainException{
		SetOfFDs clone = fds.clone();
		SetOfFDs F = new SetOfFDs();
		AttributeSet X, Y;
		FunctionalDependency fd;
		// FDs auf Form X->A bringen
		int n = clone.size();
		for (int i = 1; i <= n; i++) {
			fd = clone.first();
			X = fd.getDominator();
			Y = fd.getDominated();
			char[] attrchar = Y.toArray();
			for (int j = 0; j < attrchar.length; j++) {
				F.add(new FunctionalDependency(X,attrchar[j]));
			}
			clone.remove(clone.first());
		}
		// Test auf BCNF
		clone = F.clone();
		boolean bcnf = true;
		while (!clone.isEmpty()) {
			fd = clone.first();
			// FD ist trivial
			if (fd.getDominator().includes(fd.getDominated())) {
				clone.remove(clone.first());
			} else {
				SetOfAttributeSets keystemp = keys.clone();
				boolean correct = false;
				for (int i = 1; i <= keys.size(); i++) {
					if (fd.getDominator().includes(keystemp.first())) {
						correct = true;
						break;
					}
					keystemp.remove(keystemp.first());
				}
				if (!correct) {
					bcnf = false;
				}
				clone.remove(clone.first());
			}
		}
		if (bcnf) {
			return BCNF;
		}
		// Test auf 3NF
		clone = F.clone();
		boolean thirdnf = true;
		while (!clone.isEmpty()) {
			fd = clone.first();
			// FD ist trivial
			if (fd.getDominator().includes(fd.getDominated())) {
				clone.remove(clone.first());
			} else if (getPrimeAttributeSet().includes(fd.getDominated())){
				clone.remove(clone.first());
			} else {
				SetOfAttributeSets keystemp = keys.clone();
				boolean correct = false;
				for (int i = 1; i <= keys.size(); i++) {
					if (fd.getDominator().includes(keystemp.first())) {
						correct = true;
						break;
					}
					keystemp.remove(keystemp.first());
				}
				if (!correct) {
					thirdnf = false;
				}
				clone.remove(clone.first());
			}
		}
		if (thirdnf) {
			return ThirdNF;
		}
//		 Test auf 3NF
		clone = F.clone();
		boolean secondnf = true;
		while (!clone.isEmpty()) {
			fd = clone.first();
			// FD ist trivial
			if (fd.getDominator().includes(fd.getDominated())) {
				clone.remove(clone.first());
			} else if (getPrimeAttributeSet().includes(fd.getDominated())){
				clone.remove(clone.first());
			} else {
				SetOfAttributeSets keystemp = keys.clone();
				for (int i = 1; i <= keys.size(); i++) {
					if ((keystemp.first().includes(fd.getDominator())) && (!keystemp.first().equals(fd.getDominator()))) {
						secondnf = false;
						break;
					}
					keystemp.remove(keystemp.first());
				}
				clone.remove(clone.first());
			}
		}
		if (secondnf) {
			return SecondNF;
		}
		return FirstNF;
	}
}
