package functionalDependency.general;

import java.io.StringReader;
import java.util.Iterator;

import functionalDependency.parser.Parser;

/**
 * Die Klasse <code>RelationalDatabaseSchema</code> repraesentiert ein relationales 
 * Datenbankschema. Dieses besteht aus einer Attributmenge (<code>attributes</code>),
 * einer Menge funktionaler Abhaengigkeiten <code>fds</code> und einer Menge von
 * Relationsschemata (<code>dbSchema</code>).
 * <p>Implementiert sind neben den Konstruktoren die Methoden</p>
 * <ul>
 * <li> {@link #parseDBSchemaFromLine(String)}: Parsen eines Strings und Konstruktion
 * eines <code>RelationalDatabaseSchema</code>
 * <li> {@link #add(FunctionalDependency)}, {@link #addSchema(RelationSchema)}:
 * Hinzufuegen einer FD oder eines weiteren Relationsschema zum Datenbankschema
 * <li> verschiedene Gettermethoden, Vergleichsmethoden,
 * eine Klon- und eine Ausgabemethode
 * <li> {@link #createUniversalRelationSchema(String)}: Erstellen eines universellen
 * Datenbankschemas
 * </ul>
 * <p>Im Rahmen der Aufgabenstellung sind in dieser Klasse die Methoden</p>
 * <ul>
 * <li> {@link #convertInto3NF()}: zur abhaengigkeitserhaltenden und verlustlosen
 * Zerlegung in ein Schema in 3. Normalform (Synthesealgorithmus)
 * <b>(8 Punkte)</b>
 * <li> {@link #convertIntoBCNF()}: zur verlustlosen Zerlegung in ein Schema in BCNF
 * (Dekompositionsalgorithmus)
 * <b>(8 Punkte)</b>
 * </ul>
 * zu implementieren. Eine Beispielimplementation dieser Methoden ist gegeben und
 * kann um die beiden Methoden {@link #synthese()} und {@link #decomposition()}
 * erweitert werden.
 */
public class RelationalDatabaseSchema implements Cloneable {
	
	private SetOfRelationSchemata dbSchema;
	private AttributeSet attributes;
	private SetOfFDs fds;

	private RelationalDatabaseSchema(){
		dbSchema = null;
		attributes = null;
		fds = null;
	}
	
	public RelationalDatabaseSchema(String attr,SetOfFDs fd)throws MainException{
		dbSchema = new SetOfRelationSchemata();
		attributes = new AttributeSet(attr);
		fds = new SetOfFDs();
		for(FunctionalDependency ifd : fd){
			add(ifd);
		}
	}
	
	public RelationalDatabaseSchema(AttributeSet attr,SetOfFDs fd)throws MainException{
		dbSchema = new SetOfRelationSchemata();
		attributes = attr.clone();
		fds = new SetOfFDs();
		for(FunctionalDependency ifd : fd){
			add(ifd);
		}
	}

	static public RelationalDatabaseSchema parseDBSchemaFromLine(String line)
	throws MainException{
		Parser parser = null;
		StringReader reader;
		reader = new StringReader(line);
		parser = new Parser(reader);
		return parser.run();
	}
	
	private void add(FunctionalDependency fd) throws MainException{
		if(attributes.includes(fd.getDominator()) && attributes.includes(fd.getDominated())){
			if(!fds.add(fd))
				throw new RelationalDatabaseSchemaException(
						"fd already exists", RelationalDatabaseSchemaException.DOUBLE_FD);
		}else
			throw new RelationalDatabaseSchemaException(
					"unkown attr", RelationalDatabaseSchemaException.UNKOWN_ATTR);
	}
	
	/**
	 * Fuegt dem aktuellen relationalen Datenbankschema ein Relationsschema hinzu.
	 * @param rs das Relationsschema, das hinzugefuegt werden soll
	 * @throws MainException
	 */
	public void addSchema(RelationSchema r) throws MainException {
		dbSchema.add(r);
	}
	
	public SetOfRelationSchemata getRelations() {
		return dbSchema;
	}
	
	public AttributeSet getAttributes() {
		return attributes;
	}
	
	public SetOfFDs getFDs() {
		return fds;
	}
	
	public SetOfFDs getBase() throws MainException {
		return fds.base();
	}
	
	public boolean equals(Object o) {
		if(o instanceof RelationalDatabaseSchema)
			return equals((RelationalDatabaseSchema)o);
		return false;
	}

	public boolean equals(RelationalDatabaseSchema s) {
		if(s.attributes.equals(attributes) &&
		   s.fds.equals(fds) &&
		   s.dbSchema.equals(dbSchema))
			return true;
		return false;
	}
	
	public RelationalDatabaseSchema clone() {
		RelationalDatabaseSchema rc = new RelationalDatabaseSchema();
		rc.attributes = attributes.clone();
		rc.dbSchema = dbSchema.clone();
		rc.fds = fds.clone();
		return rc;
	}
	
	public String toString(){
		try{
			return "Attr: " + attributes + "\nFDs: " + fds + " = BASE " + fds.base()
				+ "\nRel: " + dbSchema.toString() + "\n";	
		}catch(MainException e){
			throw new RuntimeException(e);
		}
	}
	
	private RelationSchema createUniversalRelationSchema(String name)
	throws MainException {
		return new RelationSchema(name, attributes, fds);
	}

	/**
	 * Diese Methode zerlegt das relationale Datenbankschema abhaengigkeitserhaltend
	 * und verlustlosen in ein Schema in 3. Normalform. Dabei wird der
	 * Synthesealgorithmus verwendet, der in der Methode {@link #synthese()}
	 * implementiert ist.
	 * @return abhaengigkeitserhaltendes und verlustloses relationale Datenbankschema
	 * in 3. Normalform
	 * @throws MainException
	 */
	public RelationalDatabaseSchema convertInto3NF() throws MainException {
		RelationalDatabaseSchema rc = this.clone();
		rc.dbSchema = synthese();
		return rc;
	}

	/**
	 * Diese Methode zerlegt das relationale Datenbankschema verlustlos in ein Schema
	 * in Boyce-Codd-Normalform. Dabei wird der Dekompositionsalgorithmus
	 * verwendet, der in der Methode {@link #decomposition()} implementiert ist.
	 * @return verlustloses relationale Datenbankschema in Boyce-Codd-Normalform
	 * @throws MainException
	 */
	public RelationalDatabaseSchema convertIntoBCNF() throws MainException {
		RelationalDatabaseSchema rc = this.clone();
		rc.dbSchema = decomposition();
		return rc;
	}
	
//	Studentenversion
//	----------------
	
	/**
	 * Implementation des Synthesealgorithmus.
	 * Angelehnt an Algorithmus2 aus dem Praktikum.
	 * 
	 * @return eine Menge von Relationsschemata in 3. Normalform
	 * @throws MainException
	 */
	private SetOfRelationSchemata synthese() throws MainException {
		SetOfFDs G = fds.base();
		SetOfFDs clone = G.clone();
		SetOfRelationSchemata R = new SetOfRelationSchemata();
		int i = 0;
		while(!clone.isEmpty()) {
			i = i+1;
			AttributeSet X = clone.first().getAttributes();
			SetOfFDs pi = pi(X,G);
			R.add(new RelationSchema("R"+i,X,pi));
			clone.remove(clone.first());
		}
		SetOfRelationSchemata Rtest = R.clone();
		boolean haskey = false;
		while (!Rtest.isEmpty()) {
			RelationSchema Ri = Rtest.first();
			if (clone.fd_membership(G,Ri.getAttributes(),this.getAttributes())) {
				haskey = true;
			}
			Rtest.remove(Rtest.first());
		}
		if (!haskey) {
			i = i+1;
			AttributeSet X = G.key(this.getAttributes());
			R.add(new RelationSchema("R"+i,X,new SetOfFDs()));
		}
		R = removeIncludedRS(R);
		return R;
	}
	
	/**
	 * Implementation des Dekompositionsalgorithmus.
	 * Grundlage ist Algorithmus2 aus dem Praktikum.
	 * @return eine Menge von Relationsschemata in Boyce-Codd-Normalform
	 * @throws MainException
	 */
	private SetOfRelationSchemata decomposition() throws MainException {
		SetOfRelationSchemata R = dbSchema;
		SetOfFDs fdsclone = fds.clone();
		FunctionalDependency badrule = new FunctionalDependency("A","A");
		SetOfFDs badRFDs = new SetOfFDs();
		RelationSchema badR = new RelationSchema("R","A",fds);
		boolean done = false;
		boolean badrulefound = false;
		int i = 0;
		while (!done) {
			SetOfRelationSchemata clone = R.clone();
			badrulefound = false;
			while (!clone.isEmpty()) {
				RelationSchema Rtest = clone.first();
				SetOfFDs RtestFDs = piclosure(Rtest.getAttributes(), fdsclone);
				badRFDs = RtestFDs.clone();
				while (!RtestFDs.isEmpty()) {
					FunctionalDependency fdtotest = RtestFDs.first();
					if (fdtotest.getDominated().intersection(fdtotest.getDominator()).isEmpty() && !RtestFDs.coverForAttrSet(fdtotest.getDominator()).equals(Rtest.getAttributes())) {
						badrulefound = true;
						badrule = fdtotest.clone();
						break;
					}
					RtestFDs.remove(RtestFDs.first());
				}
				if (badrulefound) {
					badR = Rtest.clone();
					break;
				}
				clone.remove(clone.first());
			}
			if (badrulefound) {
				AttributeSet Xi1, Xi2;
				Xi1 = badrule.getAttributes();
				Xi2 = badR.getAttributes().substract(badrule.getDominated().substract(badrule.getDominator()));
				R.add(new RelationSchema("R"+(i+1),Xi1,piclosure(Xi1,badRFDs)));
				R.add(new RelationSchema("R"+(i+2),Xi2,piclosure(Xi2,badRFDs)));
				R.remove(badR);
				i = i+2;
			} else {
				done = true;
			}
		}
		R = removeIncludedRS(R);
		return R;
	}
	
	/**
	 * Entfernung von Teilschemata.
	 * 
	 * @param R Relationsschemata
	 * @return eine reduzierte Menge von Relationsschemata
	 * @throws MainException
	 */
	public SetOfRelationSchemata removeIncludedRS (SetOfRelationSchemata R) throws MainException {
		SetOfRelationSchemata clone = R.clone();
		SetOfRelationSchemata result = R.clone();
		AttributeSet X;
		RelationSchema Ri, Ri2;
		while (!clone.isEmpty()) {
			Ri = clone.first();
			X = Ri.getAttributes();
			clone.remove(Ri);
			SetOfRelationSchemata restlist = clone.clone();
			while (!restlist.isEmpty()) {
				Ri2 = restlist.first();
				if (Ri2.getAttributes().includes(X)) {
						result.remove(Ri);
				}
				if (X.includes(Ri2.getAttributes())) {
					result.remove(Ri2);
				}
				restlist.remove(Ri2);
			}
		}
		return result;
	}
	
	/**
	 * Gibt die Menge von FDs zurueck, die zu einer Menge von FDs gehoeren.
	 * (siehe Praktikum).
	 * 
	 * @param a Attributmenge, fd Menge von funktionalen Abhaengigkeiten
	 * @return eine reduzierte Menge von FDs
	 */
	public SetOfFDs pi(AttributeSet a, SetOfFDs fd) {
		SetOfFDs clone = fd.clone();
		SetOfFDs pi = new SetOfFDs();
		while(!clone.isEmpty()) {
			if (a.includes(clone.first().getAttributes())) {
				pi.add(clone.first());
			}
			clone.remove(clone.first());
		}
		return pi;
	}
	
	/**
	 * Gibt die Menge von FDs zurueck, die zur Huelle einer Menge von FDs
	 * gehoeren (siehe Praktikum).
	 * 
	 * @param a Attributmenge, fd Menge von funktionalen Abhaengigkeiten
	 * @return eine reduzierte Menge von FDs
	 */
	public SetOfFDs piclosure(AttributeSet a, SetOfFDs fd) throws AttributeSetException, MainException{
		SetOfFDs clone = fd.clone();
		SetOfFDs pi = new SetOfFDs();
		FunctionalDependency fd2;
		while(!clone.isEmpty()) {
			fd2 = clone.first();
			char [] attrchar = fd2.getDominated().toArray();
			AttributeSet domnew = new AttributeSet();
			if (a.includes(fd2.getDominator())) {
				for (int i = 0; i < attrchar.length;i++) {
					if (a.contains(attrchar[i])) {
						domnew.add(attrchar[i]);
					}
				}
				if (!domnew.isEmpty()) {
					pi.add(new FunctionalDependency(fd2.getDominator(),domnew));
				}
			}
			clone.remove(clone.first());
		}
		return pi;
	}
}
