package functionalDependency.general;

import java.io.StringReader;
import java.util.Iterator;
import java.util.TreeSet;

import functionalDependency.parser.Parser;

/**
 * Die Klasse <code>SetOfFDs</code> repraesentiert eine Menge von funktionalen
 * Abhaengigkeiten. Sie ist von der abstrakten Klasse {@link SetOf} abgeleitet.
 * <p>Implementiert sind neben dem Konstruktor die Methoden</p>
 * <ul>
 * <li> {@link #parseSchemaFromLine(String)}: Parsen eines Strings
 * <li> {@link #getAttributes()}: Berechnen aller Attribute in der Menge FDs
 * <li> {@link #clone()}: Erstellen einer Kopie der Menge von FDs
 * <li> {@link #equals(SetOfFDs)}: eine Vergleichsmethode
 * </ul>
 * <p>Im Rahmen der Aufgabenstellung sind in dieser Klasse die Methoden</p>
 * <ul>
 * <li> {@link #coverEquals(SetOfFDs)}: Aequivalenztest zweier Mengen von FDs
 * <b>(2 Punkte)</b>
 * <li> {@link #coverForAttrSet(AttributeSet)}: Huellenalgorithmus einer Attributmenge
 * <b>(3 Punkte)</b>
 * <li> {@link #genKeys(AttributeSet)}: Berechnen aller Schluesselkandidaten
 * <b>(5 Punkte)</b>
 * <li> {@link #base()}: Berechnen einer minimalen Ueberdeckung (Basisalgorithmus)
 * <b>(8 Punkte)</b>
 * </ul>
 * zu implementieren.
 */
public class SetOfFDs extends SetOf<FunctionalDependency> {
	
	public AttributeSet allattributes = new AttributeSet();
	
	public SetOfFDs() {
		super(";");
	}
	
	static public SetOfFDs parseSchemaFromLine(String line) throws MainException {
		Parser parser = null;
		StringReader reader;
		reader = new StringReader(line);
		parser = new Parser(reader);
		return parser.parseFDDef();
	}

	public AttributeSet getAttributes() {
		AttributeSet attr = new AttributeSet();
		for(FunctionalDependency e : setVar){
			attr = attr.union(e.getDominated()).union(e.getDominator());
		}
		return attr;
	}
	
	public SetOfFDs clone() {
		SetOfFDs rc = new SetOfFDs();
		rc.setVar = (TreeSet<FunctionalDependency>)this.setVar.clone();
		return rc;
	}
	
	public boolean exactMatch(SetOfFDs s) {
		return setVar.equals(s.setVar);		
	}
	
	/**
	 * Prueft die Gleichheit zweier <code>SetOfFDs</code>-Objekte.
	 * Ruft dazu die Methode {@link #equals(SetOfFDs)} auf.
	 * @param o ein Objekt
	 * @return <code>true</code>, wenn das Objekt <code>o</code> vom Typ
	 * <code>SetOfFDs</code> und gleich dem aktuellen
	 * <code>SetOfFDs</code>-Objekt ist; <code>false</code>, anderfalls
	 */
	public boolean equals(Object o){
		if(o instanceof SetOfFDs)
			return equals((SetOfFDs)o);
		return false;
	}

	/**
	 * Prueft die Gleichheit zweier <code>SetOfFDs</code>-Objekte.
	 * Ruft dazu die Methode {@link #coverEquals(SetOfFDs)} auf.
	 * @param s Menge von funktionalen Abhaengigkeiten
	 * @return <code>true</code>, wenn das <code>SetOfFDs</code>-Objekt
	 * gleich dem aktuellen <code>SetOfFDs</code>-Objekt ist;
	 * <code>false</code>, anderfalls
	 */
	public boolean equals(SetOfFDs s){
		return coverEquals(s);
	}
	
	
//	Studentenversion
//	----------------
	
	/**
	 * Prueft die Gleichheit zweier <code>SetOfFDs</code>-Objekte. d. h.
	 * die Aequivalenz von zwei Mengen funktionaler Abhaengigkeiten.
	 * Zwei Mengen von FDs sind aequivalent, wenn jede FD der einen Menge
	 * aus der anderen Menge abgeleitet werden kann und umgekehrt
	 * (die Mengen umhuellen sich gegenseitig).
	 * @param s Menge von funktionalen Abhaengigkeiten
	 * @return <code>true</code>, wenn beide Mengen von FDs aequivalent sind;
	 * <code>false</code>, anderfalls
	 */
	private boolean coverEquals(SetOfFDs s) {
		SetOfFDs clonethis = this.clone();
		AttributeSet X;
		AttributeSet Y;
		AttributeSet XClos;
		FunctionalDependency fd;
		while (!clonethis.isEmpty()) {
			fd = clonethis.first();
			X = fd.getDominator();
			Y = fd.getDominated();
			XClos = s.coverForAttrSet(X);
			if (!XClos.includes(Y)) {
				return false;
			}
			clonethis.remove(clonethis.first());
		}
		while (!s.isEmpty()) {
			fd = s.first();
			X = fd.getDominator();
			Y = fd.getDominated();
			XClos = this.coverForAttrSet(X);
			if (!XClos.includes(Y)) {
				return false;
			}
			s.remove(s.first());
		}
		return true;
	}
	
	/**
	 * Huellenberechnung der Attributemenge <code>x</code> bzgl. <code>this</code>
	 * (x+).
	 * @param x Attributmenge
	 * @return Huelle von <code>x</code> bzgl. <code>this</code>
	 */
	public AttributeSet coverForAttrSet(AttributeSet x) {
		SetOfFDs clone = this.clone();
		AttributeSet Y = x;
		AttributeSet S, T;
		AttributeSet result = x;
		AttributeSet resultold = new AttributeSet();
		FunctionalDependency temp;
		while (!result.equals(resultold)) {
			resultold = result.clone();
			while (!clone.isEmpty()) {
				temp = clone.first();
				S = temp.getDominator();
				T = temp.getDominated();			
				if (Y.includes(S) && !Y.includes(T)) {
					Y = Y.union(T);
				}	
				clone.remove(clone.first());
			}
			result = Y.clone();
			clone = this.clone();
		}		
		return Y;
	}
	
	/**
	 * Generiert alle Schluesselkandidaten.
	 * 
	 * @param attributes Menge von Attributen, die als Schluesselattributen
	 * in Frage kommen
	 * @return Menge von Attributmengen, die jeweils Schluesselkandidaten sind
	 * @throws MainException
	 */
	public SetOfAttributeSets genKeys(AttributeSet attributes) throws MainException {
		allattributes = attributes;
		SetOfAttributeSets result = new SetOfAttributeSets();
		SetOfAttributeSets resultold = new SetOfAttributeSets();
		result.add(attributes);
		while (!result.equals(resultold)) {
			resultold = result.clone();
			SetOfAttributeSets resulttemp = new SetOfAttributeSets();
			while (!result.isEmpty()) {
				resulttemp = resulttemp.union(genKey(result.first()));
				result.remove(result.first());
			}
			result = resulttemp.clone();
		}
		return result;
	}
	
	/**
	 * Generiert Schluesselkandidaten, die genau ein Attribut weniger haben als
	 * die Attributmenge attributes. Falls es keinen gibt, wird die Eingabe
	 * zurueckgegeben.
	 * 
	 * @param attributes Menge von Attributen, die als Schluesselattributen
	 * in Frage kommen
	 * @return Menge von Attributmengen, die jeweils Schluesselkandidaten sind
	 * @throws MainException
	 */
	public SetOfAttributeSets genKey(AttributeSet attributes) throws MainException {
		SetOfAttributeSets result = new SetOfAttributeSets();
		result.add(attributes);
		char[] attrchar = attributes.toArray();
		for (int i = 0; i < attrchar.length; i++) {
			if (coverForAttrSet(attributes.substract(attrchar[i])).equals(allattributes)) {
				result.remove(attributes);
				result.add(attributes.substract(attrchar[i]));
			}
		}		
		return result;
	}
	
	/**
	 * Berechnung eines Schluessels fuer gegebene Attributmenge
	 * bezueglich einer Menge von FDs (this).
	 * 
	 * @param attributes Attributmenge
	 * @return K Schluessel
	 * @throws AttributeSetException
	 */
	public AttributeSet key(AttributeSet attributes) throws AttributeSetException {
		AttributeSet K = attributes.clone();
		char[] attrchar = attributes.toArray();
		for (int i = 0; i < attrchar.length; i++) {
			if (fd_membership(this,K.substract(attrchar[i]),attributes)) {
				K.remove(attrchar[i]);
			}
		}
		return K;
	}
	
	/**
	 * Berechnung einer minimalen Ueberdeckung (Basisalgorithmus).
	 * Die Implementierung beruht auf Algorithmus 2.
	 * 
	 * @return Basis der funktionalen Abhaengigkeiten
	 * @throws MainException
	 */
	public SetOfFDs base() throws MainException {
		SetOfFDs clone = this.clone();
		SetOfFDs G = new SetOfFDs();
		AttributeSet X = new AttributeSet();
		AttributeSet Y = new AttributeSet();
		FunctionalDependency fd;
		// machte F (this) rechts-minimal
		int n = clone.size();
		for (int i = 1; i <= n; i++) {
			fd = clone.first();
			X = fd.getDominator();
			Y = fd.getDominated();
			char[] attrchar = Y.toArray();
			for (int j = 0; j < attrchar.length; j++) {
				G.add(new FunctionalDependency(X,attrchar[j]));
			}
			clone.remove(clone.first());
		}
		// mache G links-minimal
		int k = G.size();
		clone = G.clone();
		for (int i = 1; i <= k; i++) {
			fd = clone.first();
			X = fd.getDominator();
			Y = fd.getDominated();
			char[] attrchar = X.toArray();
			for (int j = 0; j < attrchar.length; j++) {
				AttributeSet Xtemp = X.clone();
				if ((X.size() > 1)) {
					Xtemp.remove(attrchar[j]);
					if (fd_membership(G,Xtemp,Y)) {
						X = Xtemp.clone();
						G.remove(fd);
						G.add(new FunctionalDependency(Xtemp,Y));
					}
				}
			}
			clone.remove(fd);
		}
		// entferne redundante FDs
		int knew = G.size();
		clone = G.clone();
		for (int i = 1; i <= knew; i++) {
			fd = clone.first();
			G.remove(fd);
			if (!fd_membership(G,fd.getDominator(),fd.getDominated())) {
				G.add(fd);
			}
			clone.remove(clone.first());
		}
		// Zusammenfassen von FDs mit gleicher linker Seite
		G = putleftsidestogether(G);
		return G;
	}
	
	/**
	 * Test, ob X->Z element F+
	 * Wie in Praktikum.
	 * 
	 * @input F, X, Z
	 * @return true, wenn X->Z element F+
	 */
	public boolean fd_membership(SetOfFDs F, AttributeSet X, AttributeSet Z) {
		boolean membership = false;
		AttributeSet Y = F.coverForAttrSet(X);
		if (Y.includes(Z)) {
			membership = true;
		}
		return membership;
	}
	
	/**
	 * Zusammenfassen von FDs mit gleicher linker Seite.
	 * 
	 * @input fds Menge von FDs
	 * @return reduzierte Menge von FDs
	 * @throws MainException
	 */
	public SetOfFDs putleftsidestogether (SetOfFDs fds) throws MainException {
		SetOfFDs clone = fds.clone();
		SetOfFDs result = new SetOfFDs();
		FunctionalDependency fd, fd2;
		while (!clone.isEmpty()) {
			fd = clone.first();
			AttributeSet X = fd.getDominator();
			AttributeSet Y = fd.getDominated();
			clone.remove(fd);
			result.add(fd);
			SetOfFDs restlist = clone.clone();
			while (!restlist.isEmpty()) {
				fd2 = restlist.first();
				if (X.equals(fd2.getDominator())) {
						result.remove(new FunctionalDependency (X,Y));
						clone.remove(fd2);
						Y = Y.union(fd2.getDominated());
						result.add(new FunctionalDependency (X, Y));
				}
				restlist.remove(fd2);
			}
			clone.remove(fd);
		}
		return result;
	}
}